theory Traces
imports Automata ListFuns
begin

type_synonym
   ('a,'s)pairs            =    "('a * 's) list"

type_synonym
   ('a,'s)execution        =    "'s * ('a,'s)pairs"

type_synonym
   'a trace                =    "'a list"

type_synonym
   ('a,'s)execution_module = "('a,'s)execution set * 'a signature"

type_synonym
   'a schedule_module      = "'a trace set * 'a signature"

type_synonym
   'a trace_module         = "'a trace set * 'a signature"

fun is_exec_frag :: "[('a,'s)ioa, ('a,'s)execution] => bool"
  where
  "is_exec_frag A (s_0, Nil) = True" |
  "is_exec_frag A (s_0, (a_1, s_1) # Nil) = ((s_0, a_1, s_1) \<in> trans_of A)" |
  "is_exec_frag A (s_0, (a_n, s_n) # (x # xs)) = ((snd x, a_n, s_n) \<in> trans_of A \<and> is_exec_frag A (s_0, x#xs))"
(*
  "is_exec_frag A (s_0, (a_1, s_1) # xs) = ((s_0, a_1, s_1) \<in> trans_of A \<and> is_exec_frag A (s_1, xs))"
*)
consts

   (* Executions *)
  executions    :: "('a,'s)ioa => ('a,'s)execution set"

  (* Schedules and traces *)
  filter_act    ::"('a,'s)pairs \<Rightarrow> 'a trace"
  has_schedule  :: "[('a,'s)ioa, 'a trace] => bool"
  has_trace     :: "[('a,'s)ioa, 'a trace] => bool"
  schedules     :: "('a,'s)ioa => 'a trace set"
  traces        :: "('a,'s)ioa => 'a trace set"
  mk_trace      :: "('a,'s)ioa => ('a,'s)pairs \<Rightarrow> 'a trace"

  laststate    ::"('a,'s)execution => 's"

  (* Notions of implementation *)
  ioa_implements :: "[('a,'s1)ioa, ('a,'s2)ioa] => bool"   (infixr "=<|" 12)

  (* Execution, schedule and trace modules *)
  Execs         ::  "('a,'s)ioa => ('a,'s)execution_module"
  Scheds        ::  "('a,'s)ioa => 'a schedule_module"
  Traces        ::  "('a,'s)ioa => 'a trace_module"

defs

(*  ------------------- Executions ------------------------------ *)

executions_def:
  "executions ioa \<equiv> {e. ((fst e) \<in> starts_of ioa ) \<and>
                         is_exec_frag ioa e}"

(*  ------------------- Schedules ------------------------------ *)

filter_act_def:
  "filter_act == map fst"

has_schedule_def:
  "has_schedule ioa sch \<equiv>
     (\<exists> ex \<in> executions ioa. sch = filter_act (snd ex))"

schedules_def:
  "schedules ioa == {sch. has_schedule ioa sch}"

(*  ------------------- Traces ------------------------------ *)

has_trace_def:
  "has_trace ioa tr \<equiv>
     (\<exists> sch \<in> schedules ioa.  tr = filter (\<lambda> a. a \<in> ext(ioa)) sch)"

traces_def:
  "traces ioa == {tr. has_trace ioa tr}"


mk_trace_def:
  "mk_trace ioa == \<lambda> tr.
     filter (\<lambda> a. a \<in> ext(ioa)) (filter_act tr)"

(*  ------------------- Implementation ------------------------------ *)

ioa_implements_def:
  "ioa1 =<| ioa2 ==
    (((inputs(asig_of(ioa1)) = inputs(asig_of(ioa2))) &
     (outputs(asig_of(ioa1)) = outputs(asig_of(ioa2)))) &
      traces(ioa1) <= traces(ioa2))"

(*  ------------------- Modules ------------------------------ *)

Execs_def:
  "Execs A  == (executions A, asig_of A)"

Scheds_def:
  "Scheds A == (schedules A, asig_of A)"

Traces_def:
  "Traces A == (traces A,asig_of A)"

definition is_trace_module :: "'a trace_module \<Rightarrow> bool"
  where
  "is_trace_module \<equiv> \<lambda> m . \<forall> tr \<in> fst m . forall (\<lambda> a . a \<in> externals (snd m)) tr"

lemmas [simp del] = HOL.ex_simps HOL.all_simps split_paired_Ex
declare Let_def [simp]
declaration {* fn _ => Classical.map_cs (fn cs => cs delSWrapper "split_all_tac") *}

lemmas exec_rws = executions_def is_exec_frag_def

subsection "recursive equations of operators"

(* ---------------------------------------------------------------- *)
(*                               filter_act                         *)
(* ---------------------------------------------------------------- *)

lemma filter_act_nil: "filter_act Nil = Nil"
apply (simp add: filter_act_def)
done

lemma filter_act_cons: "filter_act (x#xs) = (fst x) # filter_act xs"
apply (simp add: filter_act_def)
done

declare filter_act_nil [simp] filter_act_cons [simp]


(* ---------------------------------------------------------------- *)
(*                             mk_trace                             *)
(* ---------------------------------------------------------------- *)

lemma mk_trace_nil: "mk_trace A Nil= Nil"
apply (simp add: mk_trace_def)
done

lemma mk_trace_cons: "mk_trace A (at # xs) =     
             (if ((fst at) \<in> ext A)            
                  then (fst at) # (mk_trace A xs)     
                  else mk_trace A xs)"
apply (simp add: mk_trace_def)
done

declare mk_trace_nil [simp] mk_trace_cons [simp]

subsection "has_trace, mk_trace"

(* alternative definition of has_trace tailored for the refinement proof, as it does not 
   take the detour of schedules *)

lemma has_trace_def2: 
"has_trace A b = (\<exists> ex \<in> executions A. b = mk_trace A (snd ex))"
apply (unfold executions_def mk_trace_def has_trace_def schedules_def has_schedule_def)
apply auto
done

subsection "signatures and executions, schedules"

(* All executions of A have only actions of A. This is only true because of the 
   predicate state_trans (part of the predicate IOA): We have no dependent types.
   For executions of parallel automata this assumption is not needed, as in par_def
   this condition is included once more. (see Lemmas 1.1.1c in CompoExecs for example) *)

lemma execfrag_in_sig:
  fixes A
  assumes "is_trans_of A"
  shows "\<forall> s . is_exec_frag A (s,xs) \<longrightarrow> forall (\<lambda> a. a \<in> act A) (filter_act xs)"
proof (induct_tac xs)
  show "\<forall> s . is_exec_frag A (s,[]) \<longrightarrow> forall (\<lambda>a. a \<in> act A) (filter_act [])" by auto
next
  fix a list
  assume 1:"\<forall> s . is_exec_frag A (s, list) \<longrightarrow> forall (\<lambda>a. a \<in> act A) (filter_act list)"
  show "\<forall> s . is_exec_frag A (s, a # list) \<longrightarrow> forall (\<lambda>a. a \<in> act A) (filter_act (a # list))"
  proof -
    { fix s
      assume 2:"is_exec_frag A (s, a # list)"
      { assume 3:"list = Nil"
        with 2 have 4:"is_exec_frag A (s, list)" and  5:"(s, fst a, snd a) \<in> trans_of A" by auto (cases a, auto)
        from 4 and 1 have "forall (\<lambda>a. a \<in> act A) (filter_act list)" by auto
        moreover from 5 and `is_trans_of A` have "fst a \<in> act A" apply(auto simp add:is_trans_of_def) apply(cases a) apply simp done
        ultimately have "forall (\<lambda>a. a \<in> act A) (filter_act (a # list))" by auto
      } moreover
      { assume 3:"list \<noteq> Nil"
        from 3 obtain x xs where "list = x#xs" apply(cases list) apply force by metis
        with 2 have 4:"is_exec_frag A (s, list)" and  5:"(snd x, fst a, snd a) \<in> trans_of A" apply - apply(cases a) apply auto apply(cases a) apply auto done
        from 4 and 1 have "forall (\<lambda>a. a \<in> act A) (filter_act list)" by auto
        moreover from 5 and `is_trans_of A` have "fst a \<in> act A" apply(auto simp add:is_trans_of_def) apply(cases a) apply simp done
        ultimately have "forall (\<lambda>a. a \<in> act A) (filter_act (a # list))" by auto
      }
      ultimately have "forall (\<lambda>a. a \<in> act A) (filter_act (a # list))" by fastsimp
    } thus ?thesis by auto
  qed
qed

lemma exec_in_sig: 
  fixes A
  assumes "is_trans_of A"
  and "x:executions A"
  shows "forall (\<lambda> a . a \<in> act A) (filter_act (snd x))"
  using assms execfrag_in_sig by (cases x) (auto simp add:executions_def)

lemma scheds_in_sig: 
  fixes A
  assumes "is_trans_of A"
  and "x:schedules A"
  shows "forall (\<lambda> a . a \<in> act A) x"
  using assms and exec_in_sig by (auto simp add: schedules_def has_schedule_def)

end